| package |
package := Package name: 'SqTools'.
package paxVersion: 0;
	basicComment: ''.


package classNames
	add: #SqScanner;
	yourself.

package binaryGlobalNames: (Set new
	yourself).

package globalAliases: (Set new
	yourself).

package allResourceNames: (Set new
	yourself).

package setPrerequisites: (IdentitySet new
	add: '..\Object Arts\Dolphin\Base\Dolphin';
	yourself).

package!

"Class Definitions"!

Object subclass: #SqScanner
	instanceVariableNames: 'source mark hereChar aheadChar token tokenType currentComment buffer typeTable'
	classVariableNames: 'TypeTable'
	poolDictionaries: ''
	classInstanceVariableNames: ''!

"Global Aliases"!


"Loose Methods"!

"End of package definition"!

"Source Globals"!

"Classes"!

SqScanner guid: (GUID fromString: '{8975A1D4-3844-4EB3-831B-1D9618A71963}')!
SqScanner comment: ''!
!SqScanner categoriesForClass!Kernel-Objects! !
!SqScanner methodsFor!

advance
	| prevToken |
	prevToken := token.
	self scanToken.
	^prevToken!

checkpoint
	"Return a copy of all changeable state.  See revertToCheckpoint:"

	| result |
	result := Array new: 3.
	result at: 1 put: self clone.
	result at: 2 put: source clone.
	result at: 3 put: currentComment copy.
	^result!

initScanner
	buffer := WriteStream on: (String new: 40).
	typeTable := TypeTable!

nextLiteral
	"Same as advance, but -4 comes back as a number instead of two tokens"

	| prevToken |
	prevToken _ self advance.
	(prevToken == #- and: [token isKindOf: Number]) ifTrue: [^self advance negated].
	^prevToken!

notify: string
	"Refer to the comment in Object|notify:."

	self error: string!

offEnd: aString
	"Parser overrides this"

	^self notify: aString!

revertToCheckpoint: checkpoint
	"Revert to the state when checkpoint was made."

	| myCopy |
	myCopy := checkpoint first.
	1 to: self class instSize do: [:i | self instVarAt: i put: (myCopy instVarAt: i)].
	source := checkpoint second.
	currentComment := checkpoint third!

scan: inputStream
	"Bind the input stream, fill the character buffers and first token 
	buffer. "

	source := inputStream.
	self step.
	self step.
	self scanToken!

scanFieldNames: stringOrArray
	"Answer an Array of Strings that are the identifiers in the input string, 
	stringOrArray. If passed an Array, just answer with that Array, i.e.,  
	assume it has already been scanned."

	| strm |
	(stringOrArray isMemberOf: Array) ifTrue: [^stringOrArray].
	self scan: (ReadStream on: stringOrArray asString).
	strm := WriteStream on: (Array new: 10).
	[tokenType = #doIt] whileFalse: 
			[tokenType = #word ifTrue: [strm nextPut: token].
			self scanToken].
	^strm contents	"SqScanner new scanFieldNames: 'abc def ghi' ('abc' 'def' 'ghi' )"!

scanLitVec
	| s |
	s := WriteStream on: (Array new: 16).
	[tokenType = #rightParenthesis or: [tokenType = #doIt]] whileFalse: 
			[tokenType = #leftParenthesis 
				ifTrue: 
					[self
						scanToken;
						scanLitVec]
				ifFalse: 
					[tokenType = #word | (tokenType = #keyword) | (tokenType = #colon) 
						ifTrue: [self scanLitWord]
						ifFalse: 
							[(token == #- and: [(typeTable at: hereChar asciiValue) = #xDigit]) 
								ifTrue: 
									[self scanToken.
									token := token negated]]].
			s nextPut: token.
			self scanToken].
	token := s contents!

scanLitWord
	"Accumulate keywords and asSymbol the result."

	| t |
	[(typeTable at: hereChar asciiValue) = #xLetter] whileTrue: 
			[t := token.
			self xLetter.
			token := t , token].
	token := token asSymbol!

scanMessageParts: sourceString
	"Return an array of the form (comment keyword comment arg comment keyword comment arg comment) for the message pattern of this method.  Courtesy of Ted Kaehler, June 1999"

	| coll nonKeywords |
	coll := OrderedCollection new.
	self scan: (ReadStream on: sourceString asString).
	nonKeywords := 0.
	[tokenType = #doIt] whileFalse: 
			[(currentComment == nil or: [currentComment isEmpty]) 
				ifTrue: [coll addLast: nil]
				ifFalse: 
					[coll addLast: currentComment removeFirst.
					[currentComment isEmpty] 
						whileFalse: [coll at: coll size put: coll last , ' ' , currentComment removeFirst]].
			(token numArgs < 1 or: [token = #| & (coll size > 1)]) 
				ifTrue: 
					[(nonKeywords := nonKeywords + 1) > 1 ifTrue: [^coll]
					"done with header"]
				ifFalse: [nonKeywords := 0].
			coll addLast: token.
			self scanToken].
	(currentComment == nil or: [currentComment isEmpty]) 
		ifTrue: [coll addLast: nil]
		ifFalse: 
			[coll addLast: currentComment removeFirst.
			[currentComment isEmpty] 
				whileFalse: [coll at: coll size put: coll last , ' ' , currentComment removeFirst]].
	^coll!

scanStringStruct
	| s |
	s := WriteStream on: (Array new: 16).
	[tokenType = #rightParenthesis or: [tokenType = #doIt]] whileFalse: 
			[tokenType = #leftParenthesis 
				ifTrue: 
					[self
						scanToken;
						scanStringStruct]
				ifFalse: [tokenType = #word ifFalse: [^self error: 'only words and parens allowed']].
			s nextPut: token.
			self scanToken].
	token := s contents!

scanStringStruct: textOrString 
	"The input is a string whose elements are identifiers and parenthesized
	 groups of identifiers.  Answer an array reflecting that structure, representing
	 each identifier by an uninterned string."

	self scan: (ReadStream on: textOrString asString).
	self scanStringStruct.
	^token

	"SqScanner new scanStringStruct: 'a b (c d) (e f g)'"!

scanToken
	[(tokenType := typeTable at: hereChar asciiValue) == #xDelimiter] whileTrue: [self step].
	"Skip delimiters fast, there almost always is one."
	mark := source position - 1.
	(tokenType at: 1) = $x 
		ifTrue: 
			["x as first letter"

			self perform: tokenType
			"means perform to compute token & type"]
		ifFalse: 
			[token := self step asSymbol
			"else just unique the first char"].
	^token!

scanTokens: textOrString 
	"Answer an Array that has been tokenized as though the input text, 
	textOrString, had appeared between the array delimitors #( and ) in a 
	Smalltalk literal expression."

	self scan: (ReadStream on: textOrString asString).
	self scanLitVec.
	^token

	"SqScanner new scanTokens: 'identifier keyword: 8r31 ''string'' .'"!

step
	| c |
	c := hereChar.
	hereChar := aheadChar.
	source atEnd 
		ifTrue: 
			[aheadChar := 30 asCharacter
			"doit"]
		ifFalse: [aheadChar := source next].
	^c!

xBinary
	tokenType := #binary.
	token := Symbol internCharacter: self step.
	[(typeTable at: hereChar asciiValue) = #xBinary and: [hereChar ~= $-]] 
		whileTrue: [token := (token , (String with: self step)) asSymbol]!

xColon
	"Allow := for assignment by converting to #_"

	aheadChar = $= 
		ifTrue: 
			[self step.
			tokenType := #leftArrow.
			self step.
			^token := #_].
	"Otherwise, just do what normal scan of colon would do"
	tokenType := #colon.
	^token := self step asSymbol!

xDelimiter
	"Ignore blanks, etc."

	self scanToken!

xDigit
	"Form a number."

	tokenType := #number.
	(aheadChar = 30 asCharacter 
		and: [source atEnd and: 
					[source skip: -1.
					source next ~= 30 asCharacter]]) 
			ifTrue: 
				[source skip: -1
				"Read off the end last time"]
			ifFalse: [source skip: -2].
	token := [Number readFrom: source] on: Error do: [:err | self offEnd: err].
	self
		step;
		step!

xDollar
	"Form a Character literal."

	self step.
	"pass over $"
	token := self step.
	tokenType := #number
	"really should be Char, but rest of compiler doesn't know"!

xDoubleQuote
	"Collect a comment."

	"wod 1/10/98: Allow 'empty' comments by testing the first character 
	for $"

	"rather than blindly adding it to the comment being collected."

	| aStream stopChar |
	stopChar := 30 asCharacter.
	aStream := WriteStream on: (String new: 200).
	self step.
	[hereChar == $"] whileFalse: 
			[(hereChar == stopChar and: [source atEnd]) 
				ifTrue: [^self offEnd: 'Unmatched comment quote'].
			aStream nextPut: self step].
	self step.
	currentComment == nil 
		ifTrue: [currentComment := OrderedCollection with: aStream contents]
		ifFalse: [currentComment add: aStream contents].
	self scanToken!

xLetter
	"Form a word or keyword."

	| type |
	buffer reset.
	[(type := typeTable at: hereChar asciiValue) == #xLetter or: [type == #xDigit]] 
		whileTrue: 
			["open code step for speed"

			buffer nextPut: hereChar.
			hereChar := aheadChar.
			source atEnd 
				ifTrue: 
					[aheadChar := 30 asCharacter
					"doit"]
				ifFalse: [aheadChar := source next]].
	(type == #colon or: [type = #xColon and: [aheadChar ~= $=]]) 
		ifTrue: 
			[buffer nextPut: self step.
			tokenType := #keyword]
		ifFalse: [tokenType := #word].
	token := buffer contents!

xLitQuote
	"Symbols and vectors: #(1 (4 5) 2 3) #ifTrue:ifFalse: #'abc'."

	| start |
	self step.
	"litQuote"
	self scanToken.
	tokenType = #leftParenthesis 
		ifTrue: 
			[start := mark.
			self
				scanToken;
				scanLitVec.
			tokenType == #doIt 
				ifTrue: 
					[mark := start.
					self offEnd: 'Unmatched parenthesis']]
		ifFalse: 
			[(#(#word #keyword #colon) includes: tokenType) 
				ifTrue: [self scanLitWord]
				ifFalse: 
					[tokenType == #literal 
						ifTrue: 
							[(token isMemberOf: Symbol) 
								ifTrue: 
									["##word"

									token := token
									"May want to move toward ANSI here"]]
						ifFalse: [tokenType == #string ifTrue: [token := token asSymbol]]]].
	tokenType := #literal
	"#(Pen) 
	#Pen 
	#'Pen' 
	##Pen 
	###Pen
	"!

xSingleQuote
	"String."

	self step.
	buffer reset.
	
	[hereChar = $' and: 
			[aheadChar = $' 
				ifTrue: 
					[self step.
					false]
				ifFalse: [true]]] 
			whileFalse: 
				[buffer nextPut: self step.
				(hereChar = 30 asCharacter and: [source atEnd]) 
					ifTrue: [^self offEnd: 'Unmatched string quote']].
	self step.
	token := buffer contents.
	tokenType := #string! !
!SqScanner categoriesFor: #advance!expression types!public! !
!SqScanner categoriesFor: #checkpoint!expression types!public! !
!SqScanner categoriesFor: #initScanner!initialize/release!public! !
!SqScanner categoriesFor: #nextLiteral!expression types!public! !
!SqScanner categoriesFor: #notify:!error handling!public! !
!SqScanner categoriesFor: #offEnd:!error handling!public! !
!SqScanner categoriesFor: #revertToCheckpoint:!expression types!public! !
!SqScanner categoriesFor: #scan:!initialize/release!public! !
!SqScanner categoriesFor: #scanFieldNames:!public!public access! !
!SqScanner categoriesFor: #scanLitVec!expression types!public! !
!SqScanner categoriesFor: #scanLitWord!expression types!public! !
!SqScanner categoriesFor: #scanMessageParts:!public!public access! !
!SqScanner categoriesFor: #scanStringStruct!expression types!public! !
!SqScanner categoriesFor: #scanStringStruct:!public!public access! !
!SqScanner categoriesFor: #scanToken!expression types!public! !
!SqScanner categoriesFor: #scanTokens:!public!public access! !
!SqScanner categoriesFor: #step!expression types!public! !
!SqScanner categoriesFor: #xBinary!multi character scans!public! !
!SqScanner categoriesFor: #xColon!multi character scans!public! !
!SqScanner categoriesFor: #xDelimiter!multi character scans!public! !
!SqScanner categoriesFor: #xDigit!multi character scans!public! !
!SqScanner categoriesFor: #xDollar!multi character scans!public! !
!SqScanner categoriesFor: #xDoubleQuote!multi character scans!public! !
!SqScanner categoriesFor: #xLetter!multi character scans!public! !
!SqScanner categoriesFor: #xLitQuote!multi character scans!public! !
!SqScanner categoriesFor: #xSingleQuote!multi character scans!public! !

!SqScanner class methodsFor!

initialize
	"SqScanner initialize"

	| newTable |
	newTable := Array new: 256 withAll: #xBinary.
	"default"
	newTable atAll: #(9 10 12 13 32) put: #xDelimiter.
	"tab lf ff cr space"
	newTable atAll: ($0 asciiValue to: $9 asciiValue) put: #xDigit.
	newTable atAll: ($A asciiValue to: $Z asciiValue) put: #xLetter.
	newTable atAll: ($a asciiValue to: $z asciiValue) put: #xLetter.
	newTable at: 30 put: #doIt.
	newTable at: $" asciiValue put: #xDoubleQuote.
	newTable at: $# asciiValue put: #xLitQuote.
	newTable at: $$ asciiValue put: #xDollar.
	newTable at: $' asciiValue put: #xSingleQuote.
	newTable at: $: asciiValue put: #xColon.
	newTable at: $( asciiValue put: #leftParenthesis.
	newTable at: $) asciiValue put: #rightParenthesis.
	newTable at: $. asciiValue put: #period.
	newTable at: $; asciiValue put: #semicolon.
	newTable at: $[ asciiValue put: #leftBracket.
	newTable at: $] asciiValue put: #rightBracket.
	newTable at: ${ asciiValue put: #leftBrace.
	newTable at: $} asciiValue put: #rightBrace.
	newTable at: $^ asciiValue put: #upArrow.
	newTable at: $_ asciiValue put: #leftArrow.
	newTable at: $| asciiValue put: #verticalBar.
	TypeTable := newTable
!

isLiteralSymbol: aSymbol
	"Test whether a symbol can be stored as # followed by its characters.  
	Symbols created internally with asSymbol may not have this property,  
	e.g. '3' asSymbol."

	| i ascii type |
	i := aSymbol size.
	i = 0 ifTrue: [^false].
	i = 1 
		ifTrue: [('$''"()#0123456789' includes: (aSymbol at: 1)) ifTrue: [^false] ifFalse: [^true]].
	ascii := (aSymbol at: 1) asciiValue.
	"TypeTable should have been origined at 0 rather than 1 ..."
	ascii = 0 ifTrue: [^false].
	type := TypeTable at: ascii.
	(type == #xColon or: [type == #verticalBar]) ifTrue: [^i = 1].
	type == #xBinary 
		ifTrue: 
			[[i > 1] whileTrue: 
					[ascii := (aSymbol at: i) asciiValue.
					ascii = 0 ifTrue: [^false].
					(TypeTable at: ascii) == #xBinary ifFalse: [^false].
					i := i - 1].
			^true].
	type == #xLetter 
		ifTrue: 
			[[i > 1] whileTrue: 
					[ascii := (aSymbol at: i) asciiValue.
					ascii = 0 ifTrue: [^false].
					type := TypeTable at: ascii.
					(type == #xLetter or: [type == #xDigit or: [type == #xColon]]) ifFalse: [^false].
					i := i - 1].
			^true].
	^false!

new
	^super new initScanner! !
!SqScanner class categoriesFor: #initialize!class initialization!public! !
!SqScanner class categoriesFor: #isLiteralSymbol:!public!testing! !
!SqScanner class categoriesFor: #new!instance creation!public! !

"Binary Globals"!

"Resources"!

